import collections
from copy import deepcopy

from slickconf.config import instantiate


def run_hook(node, hooks):
    node = deepcopy(node)
    _run_hook(node, hooks)

    return node


def _run_hook(node, hooks):
    target_key = "__target"
    init_key = "__init"
    fn_key = "__fn"
    validate_key = "__validate"
    partial_key = "__partial"
    args_key = "__args"
    key_key = "__key"

    exclude_keys = {
        target_key,
        init_key,
        fn_key,
        validate_key,
        partial_key,
        args_key,
        key_key,
    }

    if isinstance(node, collections.abc.Sequence) and not isinstance(node, str):
        seq = [_run_hook(n, hooks) for i, n in enumerate(node)]

        return seq

    if isinstance(node, collections.abc.Mapping):
        if target_key in node or init_key in node:
            if init_key in node:
                target = node.get(init_key)
                fn_key = init_key

            else:
                target = node.get(target_key)
                fn_key = target_key

            for config_key, config_hook in hooks.items():
                if target == config_key:
                    if isinstance(config_hook, str):
                        node[fn_key] = config_hook
                        res = instantiate(node, conf=node)
                        node.clear()
                        node.update(res)

                    else:
                        new_node = config_hook(node)
                        node.clear()
                        node.update(new_node)

            if args_key in node:
                args_node = node[args_key]

                for arg in args_node:
                    _run_hook(arg, hooks)

            for k, v in node.items():
                if k in exclude_keys:
                    continue

                _run_hook(v, hooks)

    return node



import collections

from slickconf.constants import (
    INIT_KEY,
    FN_KEY,
    TARGET_KEY,
    KEY_KEY,
    ARGS_KEY,
    EXCLUDE_KEYS,
    ANNOTATE_KEY,
)


def _addindent(s_, numSpaces):
    s = s_.split("\n")
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(numSpaces * " ") + line for line in s]
    s = "\n".join(s)
    s = first + "\n" + s
    return s


class ListFormatter(list):
    def __repr__(self):
        list_of_reprs = [repr(item) for item in self]
        if len(list_of_reprs) == 0:
            return "[]"

        start_end_indices = [[0, 0]]
        repeated_blocks = [list_of_reprs[0]]
        for i, r in enumerate(list_of_reprs[1:], 1):
            if r == repeated_blocks[-1]:
                start_end_indices[-1][1] += 1

                continue

            start_end_indices.append([i, i])
            repeated_blocks.append(r)

        if all([start == end for start, end in start_end_indices]):
            return super().__repr__()

        lines = []
        main_str = "["
        for (start_id, end_id), b in zip(start_end_indices, repeated_blocks):
            local_repr = f"({start_id}): {b}"  # default repr

            if start_id != end_id:
                n = end_id - start_id + 1
                local_repr = f"({start_id}-{end_id}): {n} x {b}"

            local_repr = _addindent(local_repr, 2)
            lines.append(local_repr)

        main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += "]"
        return main_str


class CallFormatter:
    def __init__(self, name, args, kwargs, key, partial):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.key = key
        self.partial = partial

        if self.key is not None:
            self.key = self.key.rsplit("#", 1)[-1]

    def format_val(self, val):
        if isinstance(val, str):
            val = f'"{val}"'

        else:
            val = str(val)

        return _addindent(val, 2)

    def __repr__(self):
        args = [self.format_val(v) for v in self.args]
        kwargs = [f"{k}={self.format_val(v)}" for k, v in self.kwargs.items()]
        MAX_LEN = 79

        params = args + kwargs

        if self.partial:
            params = [self.name] + params

        if len(", ".join(params)) < MAX_LEN:
            inline = True
            params = ", ".join(params)

        else:
            inline = False
            params = ",\n  ".join(params)

        if self.partial:
            if inline:
                result = f"partial({params})"

            else:
                result = f"partial(\n  {params}\n)"

        else:
            if inline:
                result = f"{self.name}({params})"

            else:
                result = f"{self.name}(\n  {params}\n)"

        return result


class AnnotateFormatter:
    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return f"[{self.key}]"


def patch_dict(node):
    partial = False
    if INIT_KEY in node:
        target = node[INIT_KEY]

    elif FN_KEY in node:
        target = node[FN_KEY]
        partial = True

    elif TARGET_KEY in node:
        target = node[TARGET_KEY]

    elif ANNOTATE_KEY in node:
        return AnnotateFormatter(node[ANNOTATE_KEY])

    else:
        return node

    args = []
    if ARGS_KEY in node:
        args = node[ARGS_KEY]

    kwargs = {}
    for k, v in node.items():
        if k in EXCLUDE_KEYS:
            continue

        kwargs[k] = v

    key = None
    if KEY_KEY in node:
        key = node[KEY_KEY]

    return CallFormatter(target, args, kwargs, key, partial)


def summarize(node, skip_keys=()):
    if isinstance(node, collections.abc.Sequence) and not isinstance(node, str):
        return ListFormatter([summarize(n) for n in node])

    elif isinstance(node, collections.abc.Mapping):
        results = {}

        for k, v in node.items():
            if k in skip_keys:
                continue

            results[k] = summarize(v)

        results = patch_dict(results)

        return results

    else:
        return node



import ast
import builtins
import copy
import functools
import importlib
import os
import inspect
import pydoc
import textwrap
import types
import uuid
from contextlib import contextmanager
from copy import deepcopy
from typing import Tuple, Union

from slickconf.config import resolve_module
from slickconf.container import Field

CFG_PACKAGE_NAME = "slickconf._conf_loader"


def str_to_import(name: str):
    obj = pydoc.locate(name)

    if obj is None:
        obj = resolve_module(name)

    return obj


def validate_syntax(filename: str):
    with open(filename) as f:
        code = f.read()

    try:
        ast.parse(code)

    except SyntaxError as e:
        raise SyntaxError(f"{filename} has syntax error") from e


def random_package_name(filename: str):
    # generate a random package name when loading config files
    return CFG_PACKAGE_NAME + str(uuid.uuid4())[:4] + "." + os.path.basename(filename)


def import_to_str(obj: object):
    module, qualname = obj.__module__, obj.__qualname__

    module_parts = module.split(".")

    for i in range(1, len(module_parts)):
        prefix = ".".join(module_parts[:i])
        candid = f"{prefix}.{qualname}"

        try:
            if str_to_import(candid) is obj:
                return candid

        except ImportError:
            pass

    return f"{module}.{qualname}"


CALL_HANDLER_ID = "__auto_config_call_handler__"
CLOSURE_WRAPPER_ID = "__auto_config_closure_wrapper__"
EMPTY_ARGUMENTS = ast.arguments(
    posonlyargs=[], args=[], kwonlyargs=[], kw_defaults=[], defaults=[]
)


def auto_config_call_handler(fn_or_cls: callable, *args, **kwargs):
    if (
        fn_or_cls is Field
        or fn_or_cls is finalize
        or fn_or_cls is tag
        or fn_or_cls is single
        or fn_or_cls is function
        or fn_or_cls is annotate
        or isinstance(fn_or_cls, EagerCallContainer)
        or fn_or_cls is copy.copy
        or fn_or_cls is copy.deepcopy
        or inspect.isbuiltin(fn_or_cls)
        or (inspect.isclass(fn_or_cls) and fn_or_cls.__module__ == "builtins")
    ):
        return fn_or_cls(*args, **kwargs)

    if fn_or_cls is functools.partial:
        return F[args[0]](*args[1:], **kwargs)

    return single[fn_or_cls](*args, **kwargs)


class AutoConfigNodeTransformer(ast.NodeTransformer):
    def visit_Call(self, node):
        return ast.Call(
            func=ast.Name(id=CALL_HANDLER_ID, ctx=ast.Load()),
            args=[node.func, *(self.visit(arg) for arg in node.args)],
            keywords=[self.visit(keyword) for keyword in node.keywords],
        )


def _wrap_ast_for_fn_with_closure_vars(
    module: ast.Module,
    fn: types.FunctionType,
) -> ast.Module:
    """Wraps `module.body` in a function that defines closure variables for `fn`.

    If `fn` has any free variables (i.e., it's `__code__.co_freevars` is not
    empty), we want to make sure that compiling its AST (assumed to be in the body
    of `module`) will create the same set of free variables in the resulting code
    object. However, by default this won't happen, since we would be compiling
    `fn`'s AST in the absence of its original context (e.g., just compiling a
    nested function, and not the containing one).

    To work around this issue, this function wraps `module.body` in another
    `FunctionDef` that defines dummy variables corresponding to `fn`'s free
    variables. This causes the subsequent compile step to create the right set of
    free variables, and allows us to use `fn.__closure__` when creating a
    new function object via `types.FunctionType`.

    We also add <_CALL_HANDLER_ID> as a final dummy variable, and append its value
    (the call handler) to `fn.__closure__` when creating the new function object.

    Effectively, this wrapping looks like the following Python code:

        def __auto_config_closure_wrapper__():
          closure_var_1 = None
          closure_var_2 = None
          ...
          <_CALL_HANDLER_ID> = None

          def fn(...):  # Or some expression involving a lambda.
            ...  # Contains references to the closure variables.

    Args:
      module: An `ast.Module` object whose body contains the function definition
        for `fn` (e.g., as an `ast.FunctionDef` or `ast.Lambda`).
      fn: The function to create dummy closure variables for (assumed to
        correspond to the body of `module`).

    Returns:
      A new `ast.Module` containing an additional wrapper `ast.FunctionDef` that
      defines dummy closure variables.
    """
    ast_name = lambda name: ast.Name(id=name, ctx=ast.Store())
    ast_none = ast.Constant(value=None)
    closure_var_definitions = [
        ast.Assign(targets=[ast_name(var_name)], value=ast_none)
        for var_name in fn.__code__.co_freevars + (CALL_HANDLER_ID,)
    ]

    wrapper_module = ast.Module(
        body=[
            ast.FunctionDef(
                name=CLOSURE_WRAPPER_ID,
                args=EMPTY_ARGUMENTS,
                body=[
                    *closure_var_definitions,
                    *module.body,
                ],
                decorator_list=[],
            )
        ],
        type_ignores=[],
    )
    wrapper_module = ast.fix_missing_locations(wrapper_module)
    return wrapper_module


def _find_function_code(code: types.CodeType, fn_name: str):
    """Finds the code object within `code` corresponding to `fn_name`."""
    code = [
        const
        for const in code.co_consts
        if inspect.iscode(const) and const.co_name == fn_name
    ]
    assert len(code) == 1, f"Couldn't find function code for {fn_name!r}."
    return code[0]


def _unwrap_code_for_fn(code: types.CodeType, fn: types.FunctionType):
    """Unwraps `code` to find the code object for `fn`.

    This function assumes `code` is the result of compiling an `ast.Module`
    returned by `_wrap_node_for_fn_with_closure_vars`.

    Args:
      code: A code object containing code for `fn`.
      fn: The function to find a code object for within `code`.

    Returns:
      The code object corresponding to `fn`.
    """
    code = _find_function_code(code, CLOSURE_WRAPPER_ID)
    code = _find_function_code(code, fn.__name__)
    return code


def _make_closure_cell(contents):
    """Returns `types.CellType(contents)`."""
    if hasattr(types, "CellType"):
        # `types.CellType` added in Python 3.8.
        return types.CellType(contents)  # pytype: disable=wrong-arg-count
    else:
        # For earlier versions of Python, build a dummy function to get CellType.
        dummy_fn = lambda: contents
        cell_type = type(dummy_fn.__closure__[0])
        return cell_type(contents)


def config_fn(fn):
    filename = inspect.getsourcefile(fn)
    line_number = fn.__code__.co_firstlineno
    source = textwrap.dedent(inspect.getsource(fn))
    node = ast.parse(source)
    node = AutoConfigNodeTransformer().visit(node)
    node = ast.fix_missing_locations(node)
    node = ast.increment_lineno(node, line_number - 1)
    node = _wrap_ast_for_fn_with_closure_vars(node, fn)
    code = compile(node, filename, "exec")
    code = _unwrap_code_for_fn(code, fn)

    indexed_handlers = []
    closure = list(fn.__closure__ or ())
    if CALL_HANDLER_ID in code.co_freevars:
        handler_idx = code.co_freevars.index(CALL_HANDLER_ID)
        handler = _make_closure_cell(auto_config_call_handler)
        indexed_handlers.append((handler_idx, handler))

    for handler_idx, handler in sorted(indexed_handlers):
        closure.insert(handler_idx, handler)
    closure = tuple(closure)

    auto_config_fn = types.FunctionType(code, fn.__globals__, closure=closure)
    auto_config_fn.__defaults__ = fn.__defaults__
    auto_config_fn.__kwdefaults__ = fn.__kwdefaults__

    return auto_config_fn


def auto_config_from_source(source: str, filename: str):
    source = textwrap.dedent(source)
    node = ast.parse(source)
    node = AutoConfigNodeTransformer().visit(node)
    node = ast.fix_missing_locations(node)
    code = compile(node, filename, "exec")

    return code


def finalize(node, *args, **kwargs):
    node = deepcopy(node)

    if "__fn" in node:
        name = node["__fn"]
        del node["__fn"]
        node["__init"] = name

    if len(args) > 0:
        if "__args" in node:
            node["__args"] += list(args)

        else:
            node["__args"] = list(args)

    node = {**node, **kwargs}

    return node


class NodeDict(Field):
    def __copy__(self):
        new_dict = NodeDict()
        for k, v in self.items():
            if k == "__key":
                obj, _ = v.rsplit("#", 1)
                new_dict[k] = obj + f"#{single.counter}"
                single.counter += 1

            else:
                new_dict[k] = copy.deepcopy(v)

        return NodeDict(new_dict)

    def __deepcopy__(self, memo):
        new_dict = NodeDict()
        memo[id(self)] = new_dict

        for k, v in self.items():
            if k == "__key":
                obj, _ = v.rsplit("#", 1)
                v = obj + f"#{Single.counter}"
                Single.counter += 1

            new_dict[copy.deepcopy(k, memo)] = copy.deepcopy(v, memo)

        return NodeDict(new_dict)


def build(__key, __name, *args, **kwargs):
    node = {__key: __name}

    if len(args) > 0:
        node["__args"] = list(args)

    node = NodeDict({**node, **kwargs})

    return node


def build_init(__name, *args, **kwargs):
    return build("__init", __name, *args, **kwargs)


def build_fn(__name, *args, **kwargs):
    return build("__fn", __name, *args, **kwargs)


def placeholder(default=None):
    return build("__placeholder", "__placeholder", default=default)


def annotate(annotation, value):
    return build("__annotate", annotation, value=value)


class NoValue:
    def __repr__(self):
        return "slickconf.NO_VALUE"


NO_VALUE = NoValue()


def tag(tag_name, default=NO_VALUE):
    if isinstance(default, NoValue):
        return build("__tag", tag_name)

    return build("__tag", tag_name, default=default)


def function(obj, *args, **kwargs):
    filepath = None
    qualname = None

    if not isinstance(obj, str):
        filepath = obj.__code__.co_filename
        qualname = obj.__qualname__
        obj = import_to_str(obj)

    res = build_fn(obj, *args, **kwargs)

    if filepath is not None:
        res["__meta"] = {
            "filepath": filepath,
            "qualname": qualname,
            "import_pyfile": True,
        }

    return res


class Init:
    def __init__(self, name, fn=False, key=None, filepath=None):
        self.name = name
        self.fn = fn
        self.key = key
        self.filepath = filepath

    def __call__(self, *args, **kwargs):
        if self.fn:
            res = build_fn(self.name, *args, **kwargs)

            if self.filepath is not None:
                res["_meta_"] = {"filepath": self.filepath}

            return res

        res = build_init(self.name, *args, **kwargs)

        if self.key is not None:
            res["__key"] = self.key

        if self.filepath is not None:
            res["_meta_"] = {"filepath": self.filepath}

        return res


class Single:
    counter = 0

    def __getitem__(self, obj):
        fn = False

        if not isinstance(obj, str):
            obj = import_to_str(obj)

        key = f"{obj}#{Single.counter}"
        Single.counter += 1

        return Init(obj, fn, key=key)


class EagerCallContainer:
    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)


class EagerCall:
    def __getitem__(self, obj):
        return EagerCallContainer(obj)


class LazyCall:
    def __getitem__(self, obj):
        fn = False

        if isinstance(obj, tuple):
            obj, fn = obj

        if not isinstance(obj, str):
            obj = import_to_str(obj)

        return Init(obj, fn)


class LazyFn:
    def __getitem__(self, obj):
        if not isinstance(obj, str):
            obj = import_to_str(obj)

        return Init(obj, True)


@contextmanager
def patch_import():
    """
    Enhance relative import statements in config files, so that they:
    1. locate files purely based on relative location, regardless of packages.
        e.g. you can import file without having __init__
    2. do not cache modules globally; modifications of module states has no side effect
    3. support other storage system through PathManager
    4. imported dict are turned into omegaconf.DictConfig automatically
    """
    old_import = builtins.__import__

    def find_relative_file(original_file, relative_import_path, level):
        cur_file = os.path.dirname(original_file)

        for _ in range(level - 1):
            cur_file = os.path.dirname(cur_file)

        cur_name = relative_import_path.lstrip(".")

        for part in cur_name.split("."):
            cur_file = os.path.join(cur_file, part)

        # NOTE: directory import is not handled. Because then it's unclear
        # if such import should produce python module or DictConfig. This can
        # be discussed further if needed.
        if not cur_file.endswith(".py"):
            cur_file += ".py"

        if not os.path.isfile(cur_file):
            raise ImportError(
                f"cannot import name {relative_import_path} from "
                f"{original_file}: {cur_file} has to exist"
            )

        return cur_file

    def new_import(name, globals=None, locals=None, fromlist=(), level=0):
        if (
            # Only deal with relative imports inside config files
            level != 0
            and globals is not None
            and (globals.get("__package__", "") or "").startswith(CFG_PACKAGE_NAME)
        ):
            cur_file = find_relative_file(globals["__file__"], name, level)
            validate_syntax(cur_file)
            spec = importlib.machinery.ModuleSpec(
                random_package_name(cur_file), None, origin=cur_file
            )
            module = importlib.util.module_from_spec(spec)
            module.__file__ = cur_file

            with open(cur_file) as f:
                content = f.read()

            exec(compile(content, cur_file, "exec"), module.__dict__)

            # for name in fromlist:  # turn imported dict into DictConfig automatically
            #     val = _cast_to_config(module.__dict__[name])
            #     module.__dict__[name] = val

            return module

        return old_import(name, globals, locals, fromlist=fromlist, level=level)

    builtins.__import__ = new_import
    yield new_import
    builtins.__import__ = old_import


class PyConfig:
    @staticmethod
    def load(filename: str, keys: Union[None, str, Tuple[str, ...]] = None):
        """
        Load a config file.
        Args:
            filename: absolute path or relative path w.r.t. the current working directory
            keys: keys to load and return. If not given, return all keys
                (whose values are config objects) in a dict.
        """
        has_keys = keys is not None
        filename = filename.replace("/./", "/")  # redundant
        filename = os.path.abspath(filename)

        if os.path.splitext(filename)[1] not in [".py", ".yaml", ".yml"]:
            raise ValueError(f"Config file {filename} has to be a python or yaml file.")

        if filename.endswith(".py"):
            validate_syntax(filename)

            with patch_import():
                # Record the filename
                module_namespace = {
                    "__file__": filename,
                    "__package__": random_package_name(filename),
                }
                with open(filename) as f:
                    content = f.read()

                if content.strip().startswith('"use legacy"'):
                    # Compile first with filename to:
                    # 1. make filename appears in stacktrace
                    # 2. make load_rel able to find its parent's (possibly remote) location
                    exec(compile(content, filename, "exec"), module_namespace)

                else:
                    module_namespace[CALL_HANDLER_ID] = auto_config_call_handler

                    exec(auto_config_from_source(content, filename), module_namespace)

            ret = module_namespace

        ret = ret["conf"].to_dict()

        if has_keys:
            return tuple(ret[a] for a in keys)

        return ret


L = LazyCall()
F = LazyFn()
single = Single()
call = EagerCall()



import collections
import functools
import inspect
import os
import sys
import typing
from typing import Any, Callable, Optional

from pydantic import (
    BaseModel,
    StrictBool,
    StrictInt,
    StrictStr,
    ValidationError,
    create_model,
    validator,
)

try:
    from pydantic_core import core_schema

except ImportError:
    pass

from slickconf.constants import (
    EXCLUDE_KEYS,
    TARGET_KEY,
    INIT_KEY,
    FN_KEY,
    VALIDATE_KEY,
    PARTIAL_KEY,
    ARGS_KEY,
    KEY_KEY,
    PLACEHOLDER_KEY,
    META_KEY,
    TAG_KEY,
    ANNOTATE_KEY,
)

CONFIG_REGISTRY = {}
SINGLETON = {}


class Config(BaseModel):
    class Config:
        extra = "forbid"
        protected_namespaces = ()


class StrictConfig:
    extra = "forbid"
    arbitrary_types_allowed = True


class AnyConfig:
    arbitrary_types_allowed = True


class MainConfig(BaseModel):
    class Config:
        extra = "forbid"

    distributed: Optional[StrictBool] = None
    host: Optional[StrictStr] = None
    world_size: Optional[StrictInt] = None
    rank: Optional[StrictInt] = None
    local_rank: Optional[StrictInt] = None
    ckpt: Optional[StrictStr] = None


def _check_type(type_name):
    @validator("type", allow_reuse=True)
    def check_type(cls, v):
        if v != type_name:
            raise ValueError(f"Type does not match for {type_name}")

        return v

    return check_type


def make_model_from_signature(
    name: str,
    init_fn: Callable[..., Any],
    signature: inspect.Signature,
    exclude: tuple | list,
    type_name: str | None = None,
    strict: bool = True,
):
    """Create a Pydantic model from a function signature."""

    params = {}

    if type_name is not None:
        params["type"] = (StrictStr, ...)

    for k, v in signature.parameters.items():
        if k in exclude:
            continue

        if v.kind == v.VAR_POSITIONAL or v.kind == v.VAR_KEYWORD:
            strict = False

            continue

        annotation = v.annotation
        if annotation is inspect._empty:
            annotation = typing.Any

        if v.default is inspect._empty:
            params[k] = (annotation, ...)

        else:
            params[k] = (annotation, v.default)

    def _params(self):
        values = self.dict()

        if type_name is not None:
            values.pop("type")

        return values

    @functools.wraps(init_fn)
    def _init_fn(self, *args, **kwargs):
        params = self.params()
        params.update(kwargs)
        pos_replace = list(signature.parameters.keys())[: len(args)]
        for pos in pos_replace:
            params.pop(pos)

        return init_fn(*args, **params)

    validators = {"params": _params, "make": _init_fn}

    if type_name is not None:
        validators["check_type"] = _check_type(type_name)

    if strict:
        config = StrictConfig

    else:
        config = AnyConfig

    model = create_model(
        name,
        __config__=config,
        __validators__=validators,
        __module__=__name__,
        **params,
    )

    setattr(sys.modules[__name__], name, model)

    return model


def resolve_module(path: str):
    """
    Resolves a module or attribute within a module from a dotted path.

    This function attempts to import a module or attribute using a dotted path string. It starts by trying to import the
    entire path as a module. If that fails, it progressively steps back through the path, attempting to import each
    segment as a module until it finds a valid module. Then, it attempts to resolve any remaining path segments as
    attributes within the found module.

    Parameters:
    - path (str): The dotted path string to resolve.

    Returns:
    - The resolved module or attribute.

    Raises:
    - ImportError: If the module or attribute cannot be found or imported.
    """

    from importlib import import_module

    sub_path = path.split(".")
    module = None

    for i in reversed(range(len(sub_path))):
        try:
            mod = ".".join(sub_path[:i])
            module = import_module(mod)

        except (ModuleNotFoundError, ImportError):
            continue

        if module is not None:
            break

    obj = module

    for sub in sub_path[i:]:
        mod = f"{mod}.{sub}"

        if not hasattr(obj, sub):
            try:
                import_module(mod)

            except (ModuleNotFoundError, ImportError) as e:
                raise ImportError(
                    f"Encountered error: '{e}' when loading module '{path}'"
                ) from e

        obj = getattr(obj, sub)

    return obj


def resolve_module_pyfile(path: str, filepath):
    import importlib

    spec = importlib.util.spec_from_file_location(
        os.path.splitext(os.path.basename(filepath))[0], filepath
    )
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    function = getattr(module, path)

    return function


def flatten_tree(node):
    res = []

    if isinstance(node, collections.abc.Sequence) and not isinstance(node, str):
        for n in node:
            res.extend(flatten_tree(n))

        return res

    if isinstance(node, collections.abc.Mapping):
        res.append(node)

        for v in node.values():
            res.extend(flatten_tree(v))

    return res


def find_placeholder(node):
    exclude_keys = {
        TARGET_KEY,
        INIT_KEY,
        FN_KEY,
        VALIDATE_KEY,
        PARTIAL_KEY,
        ARGS_KEY,
        KEY_KEY,
    }

    placeholders = set()

    if isinstance(node, collections.abc.Sequence) and not isinstance(node, str):
        seq = [find_placeholder(n) for i, n in enumerate(node)]

        for s in seq:
            placeholders = placeholders.union(s)

        return placeholders

    if isinstance(node, collections.abc.Mapping):
        if TARGET_KEY in node or INIT_KEY in node or FN_KEY in node:
            if ARGS_KEY in node:
                args_node = node[ARGS_KEY]

                for arg in args_node:
                    placeholders = placeholders.union(find_placeholder(arg))

            for k, v in node.items():
                if k in exclude_keys:
                    continue

                try:
                    if v[PLACEHOLDER_KEY] == PLACEHOLDER_KEY:
                        placeholders.add(k)

                        continue

                except:
                    pass

                placeholders = placeholders.union(find_placeholder(v))

    return placeholders


def _instance_traverse_validate(
    node,
    recursive=True,
    instantiate=False,
    _tags_=None,
):
    partial = node.get(PARTIAL_KEY, False)
    do_validate = node.get(VALIDATE_KEY, True)

    if INIT_KEY in node:
        target = node.get(INIT_KEY)

    elif FN_KEY in node:
        target = node.get(FN_KEY)

        if len([k for k in node if k not in EXCLUDE_KEYS]) > 0:
            partial = True

        else:
            do_validate = False

    else:
        target = node.get(TARGET_KEY)

    if node.get("__meta", None) is not None and node["__meta"].get(
        "import_pyfile", False
    ):
        obj = resolve_module_pyfile(
            node["__meta"]["qualname"], node["__meta"]["filepath"]
        )

    else:
        obj = resolve_module(target)

    signature = inspect.signature(obj)

    rest = {}

    args_replaced = []
    if ARGS_KEY in node:
        for arg, k in zip(node[ARGS_KEY], signature.parameters.keys()):
            if isinstance(arg, collections.abc.Mapping) and ANNOTATE_KEY in arg:
                arg = arg["value"]

            rest[k] = arg
            args_replaced.append(k)

    for k, v in node.items():
        if k in EXCLUDE_KEYS:
            continue

        rest[k] = instance_traverse(
            v, recursive=recursive, _tags_=_tags_, instantiate=instantiate
        )

        if k in args_replaced:
            raise TypeError(f"{target} got multiple values for argument '{k}'")

    if do_validate:
        name = "instance." + target

        exclude = []

        for r_k, r_v in rest.items():
            for e_k in EXCLUDE_KEYS:
                try:
                    if not isinstance(r_v, str) and e_k in r_v:
                        exclude.append(r_k)

                        break

                except:
                    continue

            try:
                if not isinstance(r_v, str) and TAG_KEY in r_v:
                    exclude.append(r_k)

            except:
                pass

        if partial:
            rest_key = list(rest.keys())

            for k in signature.parameters.keys():
                if k not in rest_key:
                    exclude.append(k)

            model = make_model_from_signature(
                name, obj, signature, exclude, strict=False
            )

        else:
            model = make_model_from_signature(name, obj, signature, exclude)

        try:
            if len(exclude) > 0:
                exclude = set(exclude)
                model.validate({k: v for k, v in rest.items() if k not in exclude})

            else:
                model.validate(rest)

        except ValidationError as e:
            arbitrary_flag = True

            for error in e.errors():
                if error["type"] != "type_error.arbitrary_type":
                    arbitrary_flag = False

                    break

            if not arbitrary_flag:
                raise ValueError(
                    f"Validation for {target} with {v} is failed:\n{e}"
                ) from e

    for arg in args_replaced:
        del rest[arg]

    return_dict = {**node, **rest}

    return return_dict


def _instance_traverse_instantiate(
    node,
    *args,
    recursive=True,
    instantiate=False,
    keyword_args=None,
    _tags_=None,
    root=True,
    singleton_dict=None,
):
    return_fn = False
    partial = node.get(PARTIAL_KEY, False)

    if INIT_KEY in node:
        target = node.get(INIT_KEY)

    elif FN_KEY in node:
        target = node.get(FN_KEY)

        if len([k for k in node if k not in EXCLUDE_KEYS]) > 0:
            partial = True

        else:
            return_fn = True

    else:
        target = node.get(TARGET_KEY)

    if node.get(META_KEY, None) is not None and node[META_KEY].get(
        "import_pyfile", False
    ):
        obj = resolve_module_pyfile(
            node["__meta"]["qualname"], node["__meta"]["filepath"]
        )

    else:
        obj = resolve_module(target)

    signature = inspect.signature(obj)

    if KEY_KEY in node and node[KEY_KEY] in singleton_dict:
        return singleton_dict[node[KEY_KEY]]

    if ARGS_KEY in node:
        args_node = node[ARGS_KEY]

        if len(args_node) > len(args):
            args_init = []

            for a in args_node[len(args) :]:
                args_init.append(
                    instance_traverse(
                        a,
                        recursive=recursive,
                        instantiate=instantiate,
                        keyword_args=keyword_args,
                        _tags_=_tags_,
                        root=False,
                        singleton_dict=singleton_dict,
                    )
                )

            args = list(args) + args_init

    pos_replace = list(signature.parameters.keys())[: len(args)]

    kwargs = {}

    if root and keyword_args is not None:
        for k, v in keyword_args.items():
            kwargs[k] = v

    for k, v in node.items():
        if k in EXCLUDE_KEYS:
            continue

        if k in pos_replace:
            continue

        if root and keyword_args is not None and k in keyword_args:
            kwargs[k] = keyword_args[k]

            continue

        kwargs[k] = instance_traverse(
            v,
            recursive=recursive,
            instantiate=instantiate,
            keyword_args=keyword_args,
            _tags_=_tags_,
            root=False,
            singleton_dict=singleton_dict,
        )

    if return_fn:
        instance = obj

    elif partial:
        instance = functools.partial(obj, *args, **kwargs)

    else:
        instance = obj(*args, **kwargs)

    if KEY_KEY in node and node[KEY_KEY] not in SINGLETON:
        singleton_dict[node[KEY_KEY]] = instance

    return instance


def _instance_traverse_dict(
    node,
    *args,
    recursive=True,
    instantiate=False,
    keyword_args=None,
    _tags_=None,
    root=True,
    singleton_dict=None,
):
    if TARGET_KEY in node or INIT_KEY in node or FN_KEY in node:
        if instantiate:
            return _instance_traverse_instantiate(
                node,
                *args,
                recursive=recursive,
                instantiate=instantiate,
                keyword_args=keyword_args,
                _tags_=_tags_,
                root=root,
                singleton_dict=singleton_dict,
            )

        else:
            return _instance_traverse_validate(node, recursive=recursive, _tags_=_tags_)

    elif TAG_KEY in node:
        tag = node[TAG_KEY]

        if _tags_ is not None and tag in _tags_:
            return _tags_[tag]

        elif "default" in node:
            return instance_traverse(
                node["default"],
                recursive=recursive,
                instantiate=instantiate,
                keyword_args=keyword_args,
                _tags_=_tags_,
                root=False,
                singleton_dict=singleton_dict,
            )

        else:
            raise ValueError(f"Tag '{tag}' not found in _tags_")

    elif ANNOTATE_KEY in node:
        return node["value"]

    else:
        mapping = {}

        for k, v in node.items():
            mapping[k] = instance_traverse(
                v,
                recursive=recursive,
                instantiate=instantiate,
                keyword_args=keyword_args,
                _tags_=_tags_,
                root=False,
                singleton_dict=singleton_dict,
            )

        return mapping


def instance_traverse(
    node,
    *args,
    recursive=True,
    instantiate=False,
    keyword_args=None,
    _tags_=None,
    root=True,
    singleton_dict=None,
):
    if isinstance(node, collections.abc.Sequence) and not isinstance(node, str):
        seq = [
            instance_traverse(
                i,
                recursive=recursive,
                instantiate=instantiate,
                keyword_args=keyword_args,
                _tags_=_tags_,
                root=False,
                singleton_dict=singleton_dict,
            )
            for i in node
        ]

        return seq

    if isinstance(node, collections.abc.Mapping):
        return _instance_traverse_dict(
            node,
            *args,
            recursive=recursive,
            instantiate=instantiate,
            keyword_args=keyword_args,
            _tags_=_tags_,
            root=root,
            singleton_dict=singleton_dict,
        )

    else:
        return node


def init_singleton(nodes):
    key_key = "__key"

    for node in nodes:
        if key_key not in node:
            continue

        node_key = node[key_key]

        if node_key in SINGLETON:
            continue

        restrict_node = {k: v for k, v in node.items() if k != key_key}
        instance_traverse(restrict_node)
        SINGLETON[node_key] = instance_traverse(restrict_node, instantiate=True)


class Instance(dict):
    @classmethod
    def __get_pydantic_core_schema__(self, cls, source_type):
        return core_schema.no_info_after_validator_function(
            cls.validate, core_schema.dict_schema()
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        v_new = instance_traverse(v)
        instance = cls(v_new)

        return instance

    def make(self, *args, _tags_=None, **kwargs):
        # init_singleton(flatten_tree(self))
        singleton_dict = {}

        return instance_traverse(
            self,
            *args,
            instantiate=True,
            keyword_args=kwargs,
            _tags_=_tags_,
            singleton_dict=singleton_dict,
        )

    def instantiate(self, *args, **kwargs):
        return self.make(*args, **kwargs)


def instantiate(instance, *args, _tags_: dict[str, Any] = None, **kwargs):
    """
    Instantiates an instance or traverses it if already instantiated.

    This function checks if the given instance has a 'make' method and attempts to instantiate it using that method.
    If the 'make' method is not present, it falls back to traversing the instance using the 'instance_traverse' function,
    ensuring that instantiation is attempted or the instance is properly traversed with the given arguments and keyword arguments.

    Parameters:
    - instance: The instance to be instantiated or traversed.
    - *args: Variable length argument list to be passed to the instantiation or traversal function.
    - _tags_: Optional tags to be considered during instantiation or traversal.
    - **kwargs: Arbitrary keyword arguments to be passed to the instantiation or traversal function.

    Returns:
    - The instantiated or traversed instance.
    """

    try:
        return instance.make(
            *args,
            _tags_=_tags_,
            **kwargs,
        )

    except AttributeError:
        # init_singleton(flatten_tree(instance))
        singleton_dict = {}

        return instance_traverse(
            instance,
            *args,
            instantiate=True,
            keyword_args=kwargs,
            _tags_=_tags_,
            singleton_dict=singleton_dict,
        )



from collections.abc import Mapping, Sequence

try:
    from pydantic_core import core_schema

except ImportError:
    pass


class AnyConfig(dict):
    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            if isinstance(value, dict):
                value = AnyConfig(**value)
            self[key] = value

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{item}'"
            ) from e

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        try:
            del self[item]
        except KeyError as e:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{item}'"
            ) from e


def unfold_field(x):
    if isinstance(x, Sequence) and not isinstance(x, str):
        return [unfold_field(i) for i in x]

    if isinstance(x, Mapping):
        res = {}

        for k, v in x.items():
            res[k] = unfold_field(v)

        return res

    return x


class Field(dict):
    @classmethod
    def __get_pydantic_core_schema__(self, cls, source_type):
        return core_schema.no_info_after_validator_function(
            cls.validate, core_schema.dict_schema()
        )

    @classmethod
    def validate(cls, v):
        instance = cls._recursive_init_(**v)

        return instance

    @classmethod
    def _recursive_init_(cls, **kwargs):
        node = cls()

        for key, value in kwargs.items():
            if isinstance(value, Mapping):
                value = cls._recursive_init_(**value)

            node[key] = value

        return node

    def __getitem__(self, item):
        if isinstance(item, int):
            return self["__args"][item]

        return super().__getitem__(item)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self["__args"][key] = value
        else:
            super().__setitem__(key, value)

    def __getattr__(self, item):
        try:
            return self[item]

        except KeyError as e:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{item}'"
            ) from e

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        try:
            del self[item]
        except KeyError as e:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{item}'"
            ) from e

    def __repr__(self):
        return f"{self.__class__.__name__}({dict.__repr__(self)})"

    def to_dict(self):
        return unfold_field(self)


field = Field



class Linear:
    def __init__(self, in_features: int, out_features: int):
        self.in_features = in_features
        self.out_features = out_features

    def __repr__(self):
        return f"{self.__class__.__name__}({self.in_features}, {self.out_features})"


class Sequential:
    def __init__(self, *args):
        self.layers = args

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(map(repr, self.layers))})"



INIT_KEY = "__init"
FN_KEY = "__fn"
TARGET_KEY = "__target"
VALIDATE_KEY = "__validate"
PARTIAL_KEY = "__partial"
ARGS_KEY = "__args"
KEY_KEY = "__key"
PLACEHOLDER_KEY = "__placeholder"
META_KEY = "__meta"
ANNOTATE_KEY = "__annotate"
TAG_KEY = "__tag"

EXCLUDE_KEYS = {
    INIT_KEY,
    TARGET_KEY,
    FN_KEY,
    VALIDATE_KEY,
    PARTIAL_KEY,
    ARGS_KEY,
    KEY_KEY,
    META_KEY,
}



import argparse
import ast
import json
import os
import re
import sys
from argparse import Action

try:
    from pyhocon import ConfigFactory, ConfigTree
    import _jsonnet
    import torch

except ImportError:
    _jsonnet = None
    ConfigFactory = None

from slickconf.config import Config
from slickconf.container import Field


import ast
import re


def parse_expr(expr):
    left, right = expr.split("=")
    left = left.strip()

    try:
        right = ast.literal_eval(right)

    except ValueError:
        pass

    parts = left.split(".")

    return parts, right


def apply_expr(target, parts, val):
    for i, part in enumerate(parts):
        match = re.match(r"(.*?)\[(\d+)\]", part)

        if match is not None:
            key, id = match.groups()
            id = ast.literal_eval(id)

            if i == len(parts) - 1:
                getattr(target, key)[id] = val

            else:
                target = getattr(target, key)[id]

        else:
            key = part

            if isinstance(target, dict):
                if i == len(parts) - 1:
                    target[key] = val

                else:
                    target = target[key]

            else:
                if i == len(parts) - 1:
                    setattr(target, key, val)

                else:
                    target = getattr(target, key)


def read_config(config_file: str, overrides: tuple = ()):
    if config_file.endswith(".jsonnet"):
        json_str = _jsonnet.evaluate_file(config_file)
        json_obj = json.loads(json_str)
        conf = ConfigFactory.from_dict(json_obj)

    elif config_file.endswith(".py"):
        from slickconf.builder import PyConfig

        conf = PyConfig.load(config_file)

        for override in overrides:
            apply_expr(conf, *parse_expr(override))

        return conf

    else:
        conf = ConfigFactory.parse_file(config_file)

    if len(overrides) > 0:
        for override in overrides:
            conf_overrides = ConfigFactory.parse_string(override)
            conf = ConfigTree.merge_configs(conf, conf_overrides)

    return conf.as_plain_ordered_dict()


def add_preset_arguments(parser: argparse.ArgumentParser):
    parser.add_argument("--conf", type=str, required=True)
    parser.add_argument("--ckpt", type=str)

    parser = add_distributed_args(parser)

    parser.add_argument("opts", default=None, nargs=argparse.REMAINDER)

    return parser


def preset_argparser():
    parser = argparse.ArgumentParser()

    parser = add_preset_arguments(parser)

    return parser


def add_distributed_args(parser: argparse.ArgumentParser):
    parser.add_argument("--n_gpu", type=int, default=1)
    parser.add_argument("--n_machine", type=int, default=1)
    parser.add_argument("--machine_rank", type=int, default=0)

    port = 2**15 + 2**14 + hash(os.getuid() if sys.platform != "win32" else 1) % 2**14
    parser.add_argument("--dist_url", default=f"tcp://127.0.0.1:{port}")

    return parser


def build_config(overrides: tuple, config_model: Config | None = None):
    field = Field()

    for override in overrides:
        apply_expr(field, *parse_expr(override))

    if config_model is None:
        return field

    return config_model(**field)


def load_config(config: str, config_model: Config | None = None, overrides: tuple = ()):
    """
    Loads the configuration from a given source and applies any overrides.

    This function reads the configuration from the specified source, applies any provided overrides, and then
    validates the configuration against an optional configuration model. If no model is provided, a generic
    configuration model is used. The final, validated configuration is then returned.

    Parameters:
    - config: The source of the configuration to be loaded.
    - config_model: An optional Pydantic model class that the loaded configuration will be validated against. If not
                    provided, a AnyConfig that accepts any configuration is used.
    - overrides: An optional tuple of overrides that will be applied to the configuration. Each override is a
                 key-value pair where the key specifies the configuration setting to be overridden and the value
                 specifies the new value for that setting.

    Returns:
    - The loaded and validated configuration as an instance of the specified `config_model` or a AnyConfig
      object if no model was specified.
    """

    conf = read_config(config, overrides=overrides)

    if config_model is None:
        return Field._recursive_init_(**conf)

    return config_model(**conf)


def load_arg_config(config_model, show: bool = False, parser=None):
    """
    Loads configuration based on command-line arguments and applies overrides.

    This function initializes the argument parser with or without elastic specific arguments based on the `elastic` flag.
    It then parses the command-line arguments, loads the configuration from the specified source, applies any provided
    overrides, and optionally displays the configuration. The final, validated configuration is then returned along with
    the parsed arguments if the parser is not provided externally.

    Parameters:
    - config_model: The Pydantic model class that the loaded configuration will be validated against.
    - show: A boolean flag indicating whether to print the loaded configuration to stdout.
    - elastic: A boolean flag indicating whether to initialize the argument parser with elastic specific arguments.
    - parser: An optional external argument parser. If not provided, a new parser will be initialized based on the `elastic` flag.

    Returns:
    - The loaded and validated configuration as an instance of the specified `config_model`.
    - The parsed arguments, if the parser is not provided externally.
    """

    if parser is None:
        preset_parser = preset_argparser()

    else:
        preset_parser = add_preset_arguments(parser)

    args = preset_parser.parse_args()

    conf = load_config(args.conf, config_model, args.opts)
    conf.ckpt = args.ckpt

    if parser is None:
        return conf

    else:
        return conf, args



import collections
from typing import Any

from slickconf.builder import import_to_str
from slickconf.constants import INIT_KEY, FN_KEY, TARGET_KEY


def traverse(node):
    if isinstance(node, collections.abc.Sequence) and not isinstance(node, str):
        for n in node:
            yield from traverse(n)

    elif isinstance(node, collections.abc.Mapping):
        for v in node.values():
            yield from traverse(v)

        yield node


class NodeSelection:
    def __init__(self, node, fn_or_cls):
        self.node = node
        self.fn_or_cls = import_to_str(fn_or_cls)

    def __iter__(self):
        for node in traverse(self.node):
            if INIT_KEY in node:
                target = node[INIT_KEY]

            elif FN_KEY in node:
                target = node[FN_KEY]

            elif TARGET_KEY in node:
                target = node[TARGET_KEY]

            else:
                continue

            if target == self.fn_or_cls:
                yield node

    def replace(self, target):
        new_fn_or_cls = import_to_str(target)

        for node in self:
            if INIT_KEY in node:
                node[INIT_KEY] = new_fn_or_cls

            elif FN_KEY in node:
                node[FN_KEY] = new_fn_or_cls

            else:
                node[TARGET_KEY] = new_fn_or_cls

    def map(self, fn):
        for node in self:
            new_node = fn(node)
            node.clear()
            node.update(new_node)

    def set(self, **kwargs):
        for node in self:
            for key, value in kwargs.items():
                node[key] = value

    def set_args(self, args: dict[int, Any]):
        for node in self:
            if "__args" not in node:
                continue

            for k, v in args.items():
                node["__args"][k] = v

    def get(self, key):
        for node in self:
            if key in node:
                yield node[key]

    def get_args(self, index=None):
        for node in self:
            if "__args" not in node:
                continue

            if index is None:
                yield node["__args"]

            else:
                yield node["__args"][index]


class TagSelection:
    def __init__(self, node, tag):
        self.node = node
        self.tag = tag

    def __iter__(self):
        tag_key = "__tag"

        for node in traverse(self.node):
            for key, value in node.items():
                if isinstance(value, collections.abc.Mapping) and tag_key in value:
                    if value[tag_key] == self.tag:
                        yield node, key, None

                elif isinstance(value, collections.abc.Sequence):
                    for index, elem in enumerate(value):
                        if (
                            isinstance(elem, collections.abc.Mapping)
                            and tag_key in elem
                        ):
                            if elem[tag_key] == self.tag:
                                yield node, key, index

    def replace(self, value):
        for node, key, index in self:
            if index is None:
                node[key] = value

            else:
                node[key][index] = value


def select(node, fn_or_cls=None, tag=None):
    assert (
        fn_or_cls is None or tag is None
    ), "fn_or_cls and tag cannot be both specified"

    if fn_or_cls is not None:
        return NodeSelection(node, fn_or_cls)

    else:
        return TagSelection(node, tag)



from slickconf.builder import (
    F,
    L,
    placeholder,
    single,
    function,
    tag,
    call,
    config_fn,
    annotate,
)
from slickconf.container import Field, field
from slickconf.config import Config, Instance, MainConfig, instantiate
from slickconf.loader import build_config, load_arg_config, load_config
from slickconf.selector import select
from slickconf.summarize import summarize
